// This file is part of the program FRYSK.
//
// Copyright 2006, 2008, Red Hat Inc.
//
// FRYSK is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//
// FRYSK is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with FRYSK; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
// 
// In addition, as a special exception, Red Hat, Inc. gives You the
// additional right to link the code of FRYSK with code not covered
// under the GNU General Public License ("Non-GPL Code") and to
// distribute linked combinations including the two, subject to the
// limitations in this paragraph. Non-GPL Code permitted under this
// exception must only link to the code of FRYSK through those well
// defined interfaces identified in the file named EXCEPTION found in
// the source code files (the "Approved Interfaces"). The files of
// Non-GPL Code may instantiate templates or use macros or inline
// functions from the Approved Interfaces without causing the
// resulting work to be covered by the GNU General Public
// License. Only Red Hat, Inc. may make changes or additions to the
// list of Approved Interfaces. You must obey the GNU General Public
// License in all respects for all of the FRYSK code and other code
// used in conjunction with FRYSK except the Non-GPL Code covered by
// this exception. If you modify this file, you may extend this
// exception to your version of the file, but you are not obligated to
// do so. If you do not wish to provide this exception without
// modification, you must delete this exception statement from your
// version and license this file solely under the GPL without
// exception.

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "jni.hxx"

#include "jnixx/exceptions.hxx"
#include "jnixx/elements.hxx"

using namespace java::lang;

String
frysk::sys::proc::Exe::getName(jnixx::env env, jint pid) {
  // Get the name of the EXE in /proc.
  char file[FILENAME_MAX];
  if (::snprintf(file, sizeof file, "/proc/%d/exe", (int) pid)
      >= FILENAME_MAX)
    errnoException(env, errno, "snprintf: buffer overflow");

  // /proc/$$/exe contains a soft-link specifying the name of the
  // executable, possibly with "(deleted)" appended.  That link's
  // upper bound is determined by FILENAME_MAX since that is the
  // longest possible allowable file name.
  const char deleted[] = " (deleted)";
  const int maxLen = FILENAME_MAX + sizeof(deleted) + 1;
  char link[maxLen];
  int len = ::readlink (file, link, sizeof(link) - 1);
  if (len < 0 || len >= maxLen)
    errnoException(env, errno, "readlink");
  link[len] = '\0';

  // Linux's /proc/$$/exe can get screwed up in several ways.  Detect
  // each here and return null.
  if ((int) strlen(link) != len) {
    // Assume that an EXE that has somehow ended up with an embedded
    // NUL character is invalid.  This happens when the kernel screws
    // up "mv a-really-long-file $exe" leaving the updated EXE string
    // with something like "$exe<NUL>ally-long-file (deleted)".
    userException(env, "The link %s is corrupt", file);
  }

  if (strstr(link, deleted) + strlen(deleted) - link == len) {
    // Assume (possibly incorrectly) that a trailing "(deleted)"
    // always indicates a deleted file.
    link[len - strlen(deleted)] = '\0';
    userException(env, "The link %s points to the deleted file %s",
		  file, link);
  }

  if (access(link, F_OK) != 0) {
    errnoException(env, errno, "access", "file %s", link);
  }

  // Note that some kernels have a "feature" where the link can become
  // corrupted.  Just retun that, the caller needs to decide if the
  // extracted link is valid.
  return String::NewStringUTF(env, link);
}
